// $Id: nasdsim_component_client.h,v 1.12 2000/05/09 01:57:06 ozaki Exp $

#if !defined(NASDSIM_COMPONENT_CLIENT_H)
#define NASDSIM_COMPONENT_CLIENT_H

#include <hash_map>
#include <list>
#include "nasdsim_types.h"
#include "nasdsim_component_link.h"
#include "nasdsim_tracefile.h"
#include "nasdsim_event.h"
#include "nasdsim_param.h"

class event;
class tracefile;

class StreamStatus {

 private:
  bool active;
  queue <event *> pQueue;

 public:
  StreamStatus () : active (false) {};
  ~StreamStatus ();
  bool getActive (void) { return active; }
  void setActive (bool b) { active = b; }
  void enqueue (event *evt) { pQueue.push (evt); }
  event *dequeue (void);
  bool qEmpty (void) { return pQueue.empty (); }
  int size (void) { return pQueue.size (); }
  queue <event *>& getQ (void) { return pQueue; }
};

class Stats {
 private:
  
  double         avgReadRate;
  double         avgWriteRate;
  simtime_t      avgReadTime;
  simtime_t      avgWriteTime;
  unsigned int   numReads;
  unsigned int   numWrites;
  
  simtime_t totalReadTime, totalWriteTime;
  
 public:
  
  Stats ()
    : avgReadRate (0), avgWriteRate (0), avgReadTime (0), avgWriteTime (0),
    numReads (0), numWrites (0), totalReadTime (0), totalWriteTime (0) {};
  ~Stats () {};
  
  void dumpStats (ostream & output) const;
  void updateRead (simtime_t startTime, simtime_t finishTime, sim_len_t dataSize);
  void updateWrite(simtime_t startTime, simtime_t finishTime, sim_len_t dataSize);
};

class NasdStatus {
 private:
  
  sim_component_id_t id_;
  sim_bandwidth_t    lastBandwidth_;
  
  simtime_t          startTime_;
  sim_len_t          bytesAccessed_;

 public:
  
  NasdStatus (sim_component_id_t id)
    : id_(id), lastBandwidth_(0), startTime_(0), bytesAccessed_(0) {};
  
  ~NasdStatus () {};
  
  void newInterval (simtime_t currTime) { startTime_ = currTime; bytesAccessed_ = 0; }
  sim_bandwidth_t getIntervalBandwidth (simtime_t currTime) const;
  void addBytes (sim_len_t b) { bytesAccessed_ += b; }
  sim_component_id_t getId (void) { return id_; }
  void setLastBandwidth (sim_bandwidth_t bw) { lastBandwidth_ = bw; }
  sim_bandwidth_t getLastBandwidth (void) { return lastBandwidth_; }
};
 
class component_client {
 
 private:
  
  struct CreditStatus {
    int              total;
    int              inUse;
  };
  
  //------------------------------------------------------------------------------------
  ofstream *         history_dump_file_;
  ofstream *         interval_dump_file_;
  bool               dumpIntervals;
  bool               dumpHistory;
  //------------------------------------------------------------------------------------
  
  unsigned int       traceReqCnt;
  sim_component_id_t entityId;
  component_link *   link;
  tracefile *        trace;
  Stats              clientStats;
  
  typedef hash_map <sim_component_id_t, queue <event *>*> PQsHash;
  typedef PQsHash::iterator PQsHashIter;
  PQsHash pQsHash;
  
  typedef hash_map <sim_component_id_t, CreditStatus *> CreditHash;
  typedef CreditHash::iterator CreditHashIter;
  CreditHash creditHash;
  
  typedef hash_map <sim_component_id_t, NasdStatus *> NasdStatusHash;
  typedef NasdStatusHash::iterator NasdStatusHashIter;
  NasdStatusHash nasdStatusHash;
  
  typedef hash_map <sim_component_id_t, Stats *> StatsHash;
  typedef StatsHash::const_iterator StatsHashIter;
  StatsHash statsHash;
  
  // Hash table for storing partially completed read request info. Indexed by read
  // request id
  typedef hash_map <sim_request_id_t, param_read *> PartialReadHash;
  typedef PartialReadHash::iterator PartialReadHashIter;
  PartialReadHash partialReadHash;
  
  typedef hash_map <sim_stream_id_t, StreamStatus *> StreamHash;
  typedef StreamHash::iterator StreamHashIter;
  StreamHash streamHash;
  
  bool isStreamAllowed (event *, sim_stream_id_t );
  void checkNasdQueues (list<event *>*, simtime_t);
  void dumpHistoryInfo (simtime_t);
  
 public:
  
  component_client (sim_component_id_t,
		    hash_map <sim_component_id_t, int> &, 
		    char *, sim_bandwidth_t, sim_len_t, bool, bool);
  ~component_client ();
  
  component_link *getLink (void) { return link; }
  
  void incCredit (sim_component_id_t);
  
  void updateStream (sim_stream_id_t); //, sim_component_id_t);
  
  void updateWriteStats (simtime_t startTime, 
			 simtime_t finishTime, 
			 sim_component_id_t nasdId, 
			 sim_len_t dataSize);
  void updateReadStats (simtime_t startTime, 
			simtime_t finishTime, 
			sim_component_id_t nasdId, 
			sim_len_t dataSize);
  
  void checkPartialReadRequests (
				 simtime_t startTime,
				 simtime_t finishTime,
				 param_read *p
				 );
  
  unsigned int getTraceRequestCnt (void) { return traceReqCnt; }
  void dumpIntervalInfo (simtime_t);
  list <event *> * getEventList (simtime_t tmpTime, simtime_t currTime);
  
  void dump_stats (ostream &output);
  bool hasMoreEvents (void);
};

#endif // !defined(NASDSIM_COMPONENT_CLIENT_H)
