//
// Description:
//  component_nasd contains simulation parameters for a nasd and any
//  cross event simulation state. Since NASDs need to queue event,
//  component_nasd has a queue of requests. Request contain the request ID
//  and completion time of the request. This information may appear to be
//  useless, but it is helpful when verifying the correctness of the
//  simulator when a request is dequeued by a event_nasd_read_complete or
//  event_nasd_write_complete.
//
//

// $Id: nasdsim_component_nasd.h,v 1.14 2000/05/06 15:55:58 bforney Exp $

#if !defined(NASDSIM_COMPONENT_NASD_H)
#define NASDSIM_COMPONENT_NASD_H

#include <list>
#include <queue>
#include <hash_map>
#include <sys/types.h>
#include <ostream.h>
#include "nasdsim_types.h"
#include "nasdsim_component_link.h"

// struct to hold perceived bandwidth statistics. the perceived bandwidth
// is periodically computed and the stats are reset.
typedef struct perceived_s {
  simtime_t xfer_time;
  size_t bytes_xfered;
  sim_bandwidth_t perceived_bw; // the perceived bandwidth during the
  // last quanta
  sim_component_id_t nasd_id;
} perceived_t;
typedef hash_map<sim_component_id_t, perceived_t> perceived_map_t;

// Each NASDs disk bandwidth can change overtime. The following typedefs
// define the data structure to hold information
typedef struct disk_bandwidth_s {
	simtime_t time;
	sim_bandwidth_t bandwidth;
} disk_bandwidth_t;
typedef queue<disk_bandwidth_t> disk_bandwidths_t;

typedef hash_map<sim_component_id_t, block_list_t> nasd_global_map_t;
typedef nasd_global_map_t::iterator nasd_global_map_iter_t;

#if defined(NASD_LOCAL_CACHE)
class component_nasd_cache_local;
#if defined(NASD_REMOTE_CACHING)
class component_nasd_cache_global;
#endif
#endif
class component_nasd_cache;

class component_nasd {
 public:
  component_nasd(sim_component_id_t component_id,
		 disk_bandwidths_t *disk_bandwidth,
		 simtime_t avg_seek_time,
		 size_t cache_size, 
		 size_t block_size,
		 sim_bandwidth_t, 
		 sim_len_t,
		 perceived_map_t &perceived_bw,
		 int rotational_speed = 7200);
  ~component_nasd();

  simtime_t disk_io(block_list_t *block_list, bool write,
		    simtime_t start_time);

  void dump_stats(ostream &output) const;
  
  component_link *getLink (void);

  void compute_perceived_bw(void);

  void inc_request_count(void);
  void add_perceived_bw(simtime_t xfer_time, size_t bytes_xfered);
  simtime_t get_next_request_time(void) const
    { return next_request_begin_time_; }
  void set_next_request_time(simtime_t new_time);

#if defined(NASD_LOCAL_CACHE)
  block_list_t *cache_lookup_local(sim_object_t object, sim_offset_t offset,
				   sim_len_t length, bool write,
				   simtime_t &completion_time);
  simtime_t cache_add_local(block_list_t *block_list, bool write); 
  size_t cache_evict_local(size_t num_blocks, sim_object_t object,
			   sim_offset_t offset, sim_len_t length,
			   block_list_t &evicted_clean,
			   block_list_t &evicted_dirty);
  size_t get_cache_local_size(void);
#endif // defined(NASD_LOCAL_CACHE)
#if defined(NASD_REMOTE_CACHING)
  perceived_map_t get_perceived_bw_cpy(void)
    { return perceived_bw_; }

  // global cache remote table operations
  nasd_global_map_t *cache_lookup_remote(block_list_t *block_list,
					 bool write,
					 simtime_t &completion_time);
  void cache_add_remote(block_list_t *block_list,
			sim_component_id_t remote_nasd_id);
  void cache_remove_remote(block_list_t &block_list);
  void cache_remove_remote(sim_object_t object, sim_offset_t offset,
			   sim_len_t length);
  void cache_stats_incr_returned(int ret) { stats_remote_returned_ += ret; }

  // global cache (on this nasd) operations
  block_list_t *cache_lookup_global(block_list_t *block_list,
				    sim_component_id_t owner_id,
				    simtime_t &completion_time);
  void cache_add_global(block_list_t *block_list, sim_component_id_t owner_id);
  void cache_invalidate_global(block_list_t *block_list,
			       sim_component_id_t owner_id);
  int cache_evict_global(int blocks, sim_component_id_t owner_id);

  size_t get_cache_global_num_entries(sim_component_id_t id);
  size_t get_cache_global_size(void);
  size_t get_cache_size(void);
#endif // defined(NASD_REMOTE_CACHING)
#if defined(NASD_LOCAL_CACHE) || defined(NASD_REMOTE_CACHING)
  bool is_cache_full(void);
  size_t get_max_cache_size(void) const { return cache_size_; }
  size_t get_block_size(void) const { return block_size_; }
#endif // defined(NASD_LOCAL_CACHE) || defined(NASD_REMOTE_CACHING)

 private:
  typedef struct request_s {
    sim_id_t id;
    simtime_t completion_time;
  } request_t;
#if defined(NASD_REMOTE_CACHING)
  typedef hash_map<sim_offset_t, sim_component_id_t> block_cachers_t;
  typedef block_cachers_t::iterator block_cachers_iter_t;
  typedef hash_map<sim_object_t, block_cachers_t> remote_cachers_t;
  typedef remote_cachers_t::iterator remote_cachers_iter_t;
#endif

 private:
  queue<request_t> request_queue_; // queue of pending requests

  disk_bandwidths_t *disk_bandwidths_;   // bandwidths of the disk over time
  sim_bandwidth_t cur_disk_bandwidth_;  // current disk bandwidth
  simtime_t avg_seek_time_;          // average seek time
  size_t block_size_;                // size of a disk/cache block in bytes
  int rot_speed_;

  sim_object_t prev_object_id_;
  sim_offset_t prev_offset_;

  simtime_t next_request_begin_time_; // time the next pending request can
                                      // begin
  sim_component_id_t component_id_;
  component_link *link;

  // statisitics
  long stats_total_requests_;
  perceived_map_t &perceived_bw_;

#if defined(NASD_LOCAL_CACHE)
  size_t cache_size_;
  component_nasd_cache_local *cache_local_;
#if defined(NASD_REMOTE_CACHING)
  component_nasd_cache_global *cache_global_;
  remote_cachers_t remote_cachers_tbl_;
  long stats_remote_hits_;
  long stats_remote_misses_;
  long stats_remote_returned_;
  long stats_remote_adds_;
  long stats_remote_removes_;
#endif // defined(NASD_REMOTE_CACHING)
#endif // defined(NASD_LOCAL_CACHE)
};

// *********************

#if defined(NASD_LOCAL_CACHE)
class component_nasd_cache_local {
 public:
  component_nasd_cache_local(size_t block_size);
  ~component_nasd_cache_local() {}

  block_list_t *lookup(sim_object_t object, sim_offset_t offset,
		       sim_len_t length, bool write,
		       simtime_t &completion_time);
  size_t get_size(void) { return (entries_clean_.size() +
				  entries_dirty_.size())
			    * block_size_; }
  simtime_t add(block_list_t *block_list, bool write);
  size_t evict(size_t num_blocks, sim_object_t object, sim_offset_t offset,
	       sim_len_t length, block_list_t &evicted_clean,
	       block_list_t &evicted_dirty);
  void dump_stats(ostream &output);

 private:
  typedef struct cache_entry_s {
    sim_object_t object;
    sim_offset_t offset_tag;
  } cache_entry_t;
  typedef list<cache_entry_t> cache_list_t;
  typedef cache_list_t::iterator cache_list_iter_t;

  typedef struct offset_entry_s {
    cache_list_t *list;
    cache_list_iter_t element;
  } offset_entry_t;
  typedef hash_map<sim_offset_t, offset_entry_t> offset_map_t;
  typedef offset_map_t::iterator offset_map_iter_t;
  typedef hash_map<sim_object_t, offset_map_t> object_map_t;
  typedef object_map_t::iterator object_map_iter_t;

 private:
  cache_list_t entries_clean_;
  cache_list_t entries_dirty_;
  object_map_t object_map_;
  size_t block_size_;
  sim_offset_t offset_tag_mask_;

  // statistics
  long stats_misses_;
  long stats_hits_;
  long stats_adds_;
  long stats_evicts_;
};

#if defined(NASD_REMOTE_CACHING)
class component_nasd_cache_global {
 public:
  component_nasd_cache_global(size_t block_size);

  block_list_t *lookup(block_list_t *block_list,
		      sim_component_id_t owner_nasd_id,
		      simtime_t &completion_time);
  void add( block_list_t *block_list, sim_component_id_t owner_nasd_id);
  void invalidate(block_list_t *block_list, sim_component_id_t owner_nasd_id);

  size_t get_size(void);
  int num_entires(sim_component_id_t nasd_id);
  int evict(int blocks, sim_component_id_t remote_nasd_id);
  
  void dump_stats(ostream &output);

 private:
  class cache_entry_t {
  public:
    sim_object_t object;
    sim_offset_t offset_tag;

    bool operator==(const cache_entry_t &y) const {
      return (object == y.object) && (offset_tag == y.offset_tag);
    }
  };
  typedef list<cache_entry_t> cache_list_t;
  typedef cache_list_t::iterator cache_list_iter_t;

  typedef hash_map<sim_component_id_t, cache_list_t> table_t;
  typedef table_t::iterator table_iter_t;

 private:
  table_t table_;

  size_t block_size_;
  sim_offset_t offset_tag_mask_;

  // statistics
  long stats_misses_;
  long stats_hits_;
  long stats_invalidations_;
  long stats_adds_;
  long stats_evicts_;
};
#endif // defined(NASD_REMOTE_CACHING)
#endif // defined(NASD_LOCAL_CACHE)

#endif // !defined(NASDSIM_COMPONENT_NASDS_H)
