//
// Class:
//  event_nasd, event_nasd_read, event_nasd_read_complete, event_nasd_write,
//  event_nasd_write_complete
//
// Description:
//  [see corresponding header file]
//  An event_nasd class is abstract. event_nasd_read is concrete and upon
//  simulation produces an event_nasd_read_complete. Correspondingly,
//  event_nasd_write is concrete and produces an event_nasd_write_complete
//  when simulated.
//

// $Id: nasdsim_event_nasd.C,v 1.21 2000/05/12 13:56:03 bforney Exp $

#include <algorithm>
#include "nasdsim_event_nasd.h"
#include "nasdsim_event_READ.h"
#include "nasdsim_event_WRITE.h"
#include "nasdsim_event_N2N.h"
#include "nasdsim_debug.h"

#if defined(NASD_REMOTE_CACHING)
static void
create_sorted_perceived(
			perceived_map_t *perceived,
			sim_component_id_t nasd_id,
			vector<perceived_t> &faster
			);
#endif // defined(NASD_REMOTE_CACHING)

//*****
//
// event_nasd
//

//
// event_nasd::event_nasd() [constructor]
//
// description:
//  construct a C++ object with information about the NASD object
//
// parameters:
//  copy_event - event from which to clone this event. (Useful for passing
//               common data, like the injection time,  between events.)
//  start_time - time this event should start
//
event_nasd::event_nasd(
                       const event *copy_event,
                       simtime_t start_time
                       ) : 
  event(copy_event, start_time)
{
}

//
// event_nasd::~event_nasd() [destructor]
//
// description:
//  null destructor
//
// parameters:
//  none
//
event_nasd::~event_nasd()
{
}

                       
//*****
//
// event_nasd_read
//

//
// event_nasd_read::event_nasd_read() [constructor]
//
// description:
//  construct an event_nasd_read object copying the base class event's
//  member data
//
// parameters:
//  copy_event - event from which to clone this event. (Useful for passing
//               common data, like the injection time,  between events.)
//  time - start time for the event
//  read_parameters - parameters for the request
//
event_nasd_read::event_nasd_read(
                                 const event *copy_event,
				 simtime_t time,
				 param_read *read_parameters
				 ) :
  event_nasd(copy_event, time)
{
  parameter_ = read_parameters;
}

//
// Event_nasd_read::~event_nasd_read() [destructor]
//
// description:
//  null destructor
//
// parameters:
//  none
//
event_nasd_read::~event_nasd_read()
{
}

//
// event_nasd_read::simulate()
//
// description:
//  simulate this event with the given parameters (from the comp parameter)
//
// parameters:
//  comp - a component object that contains all of the simulation parameters
//         for all components
//
// return values:
//  a pointer to a list of new events
//
#if defined(NASD_LOCAL_CACHE) && !defined(NASD_REMOTE_CACHING)
events_t *
event_nasd_read::simulate(
			  components &comp
			  )
{
  event *new_event;
  component_nasd *my_nasd;
  simtime_t completion_time;
  events_t *events = new events_t;
  block_list_t *block_list_local;
  size_t cache_size;

  my_nasd = comp.get_component_nasd(parameter_->get_nasd_id());

  completion_time = start_time_;
  // lookup the request in the local cache
  block_list_local =
    my_nasd->cache_lookup_local(parameter_->get_object(),
				parameter_->get_offset(),
				parameter_->get_length(),
				false, completion_time);

  if (block_list_local != NULL && block_list_local->size() != 0) {
    // read in the data that missed in the cache(s)
    completion_time += my_nasd->disk_io(block_list_local, false,
					completion_time);

    //
    // handle filling the cache
    //

    // winnow the block list down to the size of the cache if larger
    // than the cache.
    size_t block_size = my_nasd->get_block_size();
    size_t max_cache_size = my_nasd->get_max_cache_size();
    while ((block_list_local->size() * block_size) > max_cache_size) {
      block_list_local->pop_front(); // remove the oldest first
    }
    cache_size = my_nasd->get_cache_local_size();
    if (cache_size + block_list_local->size() * block_size > max_cache_size
	&& cache_size != 0) {
      block_list_t evicted_clean, evicted_dirty;

      // evict any entries based on the local disk reads
      my_nasd->cache_evict_local((cache_size + block_list_local->size() *
				  block_size - max_cache_size)/block_size,
				 parameter_->get_object(),
				 parameter_->get_offset(),
				 parameter_->get_length(),
				 evicted_clean,
				 evicted_dirty);

      // write out these entries
      if (!evicted_dirty.empty()) {
	completion_time += my_nasd->disk_io(&evicted_dirty, false,
					    completion_time);
      }
    }

    // add new entries to the cache
    completion_time += my_nasd->cache_add_local(block_list_local, false);
  }

  delete block_list_local;

  // update the perceived
  my_nasd->add_perceived_bw(completion_time - start_time_,
			    parameter_->get_length());
  // update request count
  my_nasd->inc_request_count();

  // update next request time
  my_nasd->set_next_request_time(completion_time);

  // generate response to request
  new_event = new event_READ_nasd_send_to_switch(dynamic_cast<event *>(this),
						 completion_time,
						 parameter_);
  events->push(new_event);

  return events;
}

#elif defined(NASD_LOCAL_CACHE) && defined(NASD_REMOTE_CACHING)

events_t *
event_nasd_read::simulate(
			  components &comp
			  )
{
  event *new_event;
  component_nasd *my_nasd;
  simtime_t completion_time;
  events_t *events = new events_t;
  block_list_t *block_list_local; // blocks not in the local cache
  nasd_global_map_t *global_map; // blocks that are in the global cache
  param_nasd_2_nasd *param;
  bool local_response_needed;
  sim_bandwidth_t my_bandwidth;
  int num_blocks_local_sat, num_blocks_disk_sat;

  my_nasd = comp.get_component_nasd(parameter_->get_nasd_id());

  my_bandwidth = my_nasd->get_perceived_bw_cpy()[parameter_->get_nasd_id()].perceived_bw;

  size_t block_size = my_nasd->get_block_size();
  size_t max_cache_size = my_nasd->get_max_cache_size();

  // the simulator does not support requests that have offsets that are not on block aligned or
  // lengths that are block multiples
  ASSERT(parameter_->get_offset()%block_size == 0);
  ASSERT(parameter_->get_length()%block_size == 0);

  completion_time = start_time_;

  // if the request was returned to us, remove the entry/entries from
  // the remote table (if a race condition occurs where another request
  // adds to this table, well no big deal from a correctness standpoint,
  // but performance will suffer.)
  if (parameter_->get_returned()) {
    my_nasd->cache_remove_remote(parameter_->get_object(),
				 parameter_->get_offset(),
				 parameter_->get_length());
    my_nasd->cache_stats_incr_returned(parameter_->get_length()/block_size);
  }

  // lookup the request in the local cache
  block_list_local =
    my_nasd->cache_lookup_local(parameter_->get_object(),
				parameter_->get_offset(),
				parameter_->get_length(),
				false, completion_time);

  // xxx this code doesn't handle requests that are not on byte boundaries
  //     probably the whole simulator will fall apart.
  if (block_list_local != NULL) {
    num_blocks_local_sat =
      (parameter_->get_length()/block_size) -
      block_list_local->size();
  } else {
    num_blocks_local_sat =
      parameter_->get_length()/my_nasd->get_block_size();
  }

  // if the local cache did not satisfy the request look it up in the
  // global cache
  if (block_list_local != NULL) {

    global_map =
      my_nasd->cache_lookup_remote(block_list_local, false, completion_time);

    // generate the events for global reads
    if (global_map != NULL) {
      if ((num_blocks_local_sat > 0 && global_map->size() > 0) ||
	  (num_blocks_local_sat == 0 && global_map->size() > 1)) {
	parameter_->set_partial(true);
	// set the length parameter for the response from this NASD.
	// it should just be the number of blocks that hit in the cache
	// (num_blocks_local_sat) plus the number that will come from disk
	// block_list_local->size().
	// this should be set again in the N2N event
	parameter_->set_length((num_blocks_local_sat +
				block_list_local->size()) * block_size); 
      }

      for (nasd_global_map_iter_t iter = global_map->begin();
	   iter != global_map->end();
	   iter++) {
	param = new param_read_forward_request(parameter_->get_nasd_id(),
					       iter->first,
					       iter->second.size() * sizeof(block_info_t),
					       new param_read(*parameter_),
					       new block_list_t(iter->second));
	new_event = new event_N2N_nasd_send_to_switch(completion_time,
						      injection_time_,
						      param);
	events->push(new_event);
      }

      delete global_map;
    }
  }

  // check if the local cache satified any of the request
  local_response_needed = num_blocks_local_sat > 0 ||
    block_list_local == NULL || !block_list_local->empty();

  num_blocks_disk_sat =
    (block_list_local == NULL) ? (parameter_->get_length() / block_size) :
    block_list_local->size();

  // xxx make sure true's and false's are correct for write field of methods

  if (block_list_local != NULL && block_list_local->size() != 0) {

    //
    // handle filling the cache
    //
#if 0
    parameter_->set_length(block_list_local->size() * block_size);
#endif
    if ((my_nasd->get_cache_size() + block_list_local->size() * block_size) <=
	max_cache_size) { // cache has enough room

      // read in the data that missed in the cache(s)
      completion_time += my_nasd->disk_io(block_list_local, false,
					completion_time);

      // add new entries to the cache
      completion_time += my_nasd->cache_add_local(block_list_local, false);

    } else {
      // cache needs eviction

      block_list_t evicted_clean, evicted_dirty;
      int blocks_left = block_list_local->size();
      perceived_map_t perceived_bw_cpy = my_nasd->get_perceived_bw_cpy();
      vector<perceived_t> faster_nasds;
      sim_component_id_t local_eviction_nasd_dest = -1;
      
      // read in the data that missed in the cache(s)
      completion_time += my_nasd->disk_io(block_list_local, false,
					  completion_time);
      
      // make sure we don't have more data than the local and global cache
      // can possibly hold
      while ((block_list_local->size() * block_size) >
	     max_cache_size * perceived_bw_cpy.size()) {
#if defined(DEBUG)
	cout << "removing: block = " << block_list_local->front().block_id_;
#endif
	block_list_local->pop_front(); // remove the oldest first
      }

      // find the highest bandwidth NASD
      create_sorted_perceived(&perceived_bw_cpy, parameter_->get_nasd_id(),
			      faster_nasds);
ASSERT(faster_nasds.size() > 0);

      //
      // Evict from "local" cache all that needs to be evicted
      // this cache includes both entries from the private cache and the
      // global cache
      //
      do {
	if (faster_nasds.back().nasd_id == parameter_->get_nasd_id()) {
	  // handle eviction on local
	  // evict any entries based on the local disk reads
	  blocks_left -=
	    my_nasd->cache_evict_local(blocks_left,
				       parameter_->get_object(),
				       parameter_->get_offset(),
				       parameter_->get_length(),
				       evicted_clean,
				       evicted_dirty);
	  // write out these entries
	  if (!evicted_dirty.empty()) {
	    completion_time += my_nasd->disk_io(&evicted_dirty, true,
						completion_time);
	  }

	  // send these entries to a faster NASD
	  perceived_bw_cpy = my_nasd->get_perceived_bw_cpy();
	  vector<perceived_t> faster_nasds_tmp;
	  create_sorted_perceived(&perceived_bw_cpy, parameter_->get_nasd_id(),
				  faster_nasds_tmp);
	  evicted_clean.splice(evicted_clean.begin(), evicted_dirty);
	  if (faster_nasds_tmp.back().nasd_id != parameter_->get_nasd_id() &&
	      fabs(faster_nasds_tmp.back().perceived_bw - my_bandwidth) > 5 && // changed this to fix oscillation
	      !evicted_clean.empty()) {
	    local_eviction_nasd_dest = faster_nasds_tmp.back().nasd_id;

	    // add these blocks to remote table for future lookup
	    my_nasd->cache_add_remote(&evicted_clean,
				      faster_nasds_tmp.back().nasd_id);

	    param =
	      new param_parent_eviction_request(parameter_->get_nasd_id(),
						faster_nasds_tmp.back().nasd_id,
						evicted_clean.size() * block_size,
						new block_list_t(evicted_clean));
	    
	    // send these blocks to the global cache if a faster NASD is found
	    new_event = new event_N2N_nasd_send_to_switch(completion_time,
							  injection_time_,
							  param);
	    events->push(new_event);
	  }
	} else {
	  blocks_left =
	    my_nasd->cache_evict_global(blocks_left,
					faster_nasds.back().nasd_id);
	}

	faster_nasds.pop_back();
      } while (!faster_nasds.empty() && blocks_left > 0);
      
      //
      // find homes in the global cache for any of the blocks that do not
      // fit in the local cache
      //

      // ensure we don't add more to the cache than we can
      while (block_list_local->size() * block_size + my_nasd->get_cache_size()
             > my_nasd->get_max_cache_size()) {
        block_list_local->pop_front();
      } 
      
      // fill the local cache
      if (block_list_local->size() < (max_cache_size/block_size)) {
	completion_time += my_nasd->cache_add_local(block_list_local, false);
	blocks_left = 0;
      } else {
	size_t copied_num;
	block_list_t *tmp = new block_list_t;
	
	for (copied_num = 0;
	     copied_num < (max_cache_size/block_size); copied_num++) {
	  tmp->push_back(block_list_local->front());
	  block_list_local->pop_front();
	}
	
	completion_time += my_nasd->cache_add_local(tmp, false);
	
	blocks_left -= max_cache_size/block_size;
      }

      //
      // find homes for these evicted blocks
      //
      // find the highest bandwidth NASD that has cache entries
      perceived_bw_cpy = my_nasd->get_perceived_bw_cpy();
      create_sorted_perceived(&perceived_bw_cpy, parameter_->get_nasd_id(),
			      faster_nasds);
      while (blocks_left > 0 && !faster_nasds.empty()) {
	if (faster_nasds.back().nasd_id != parameter_->get_nasd_id() &&
	    fabs(faster_nasds.back().perceived_bw - my_bandwidth) > 5 && // changed this to fix oscillation
	    faster_nasds.back().nasd_id != local_eviction_nasd_dest) {
	  block_list_t *remote_blocks = new block_list_t;
	  
	  ASSERT(block_list_local->size() == blocks_left);
	  
	  // add the blocks to remote_blocks
	  while (remote_blocks->size() < (max_cache_size / block_size) &&
		 blocks_left > 0) {
	    remote_blocks->push_back(block_list_local->front());
	    block_list_local->pop_front();
	    blocks_left--;
	  }
	  
	  // add these blocks to remote table for future lookup
	  my_nasd->cache_add_remote(remote_blocks,
				    faster_nasds.back().nasd_id);

	  param =
	    new param_parent_eviction_request(parameter_->get_nasd_id(),
					      faster_nasds.back().nasd_id,
					      remote_blocks->size() *
					      block_size,
					      remote_blocks);
	  
	  // send these blocks to the global cache if a faster NASD is found
	  new_event = new event_N2N_nasd_send_to_switch(completion_time,
							injection_time_,
							param);
	  events->push(new_event);
	}

	faster_nasds.pop_back();
      }
    }
  }

  delete block_list_local;

  if (local_response_needed) {
    // generate response to request
    parameter_->set_length((num_blocks_local_sat + num_blocks_disk_sat) *
			   block_size);
    new_event =
      new event_READ_nasd_send_to_switch(dynamic_cast<event *>(this),
					 completion_time, parameter_);
    events->push(new_event);
  }

  // update the perceived
  my_nasd->add_perceived_bw(completion_time - start_time_,
			    parameter_->get_length());
  // update request count
  if (!parameter_->get_partial()) {
    my_nasd->inc_request_count();
  }

  // update next request time
  my_nasd->set_next_request_time(completion_time);

#if defined(DEBUG)
  cout << "Request took " << completion_time - start_time_ << " seconds" << endl;
  cout << "  for " << parameter_->get_length() << " bytes" << endl;
#endif

  return events;
}

struct less_perceived :
  public binary_function<perceived_t, perceived_t, bool> {
  bool operator()(const perceived_t& x,
		  const perceived_t& y)
    { return x.perceived_bw < y.perceived_bw; }
};

bool operator==(
		const perceived_t &x,
		const perceived_t &y
		)
{
  return x.xfer_time == y.xfer_time && x.bytes_xfered == y.bytes_xfered &&
    x.perceived_bw == y.perceived_bw && x.nasd_id == y.nasd_id;
}

static void
create_sorted_perceived(
			perceived_map_t *perceived,
			sim_component_id_t nasd_id,
			vector<perceived_t> &faster
			)
{
  vector<perceived_t>::iterator f_iter;
  perceived_map_t::iterator p_iter;

  faster.clear();

  // add all of the perceived entires into the vector
  for (p_iter = perceived->begin(); p_iter != perceived->end(); p_iter++) {
    faster.push_back(p_iter->second);
  }

  // sort in descending order
  sort(faster.begin(), faster.end(), less_perceived());

  // remove slower entries than nasd_id
  for (f_iter = faster.begin();
       f_iter != faster.end() && f_iter->nasd_id != nasd_id; f_iter++);

  if (f_iter != faster.begin()) {
    f_iter--;
  }

  if (f_iter != faster.begin()) {
    faster.erase(faster.begin(), f_iter);
  }
}

#endif // defined(NASD_LOCAL_CACHE)

#if defined(DEBUG)
//
// event_nasd_read::print()
//
// description:
//  When DEBUG is enabled, this method display information about the event.
//
// parameters:
//  output - the ostream for display
//
// return values:
//  none
// 
void
event_nasd_read::print(
		       ostream &output
		       ) const
{
  output << "nasd_read: { start = " << start_time_ << " ";
  parameter_->print(output);
  output << " }";
}
#endif // defined(DEBUG)

//*****
//
// event_nasd_write
//

//
// event_nasd_write::event_nasd_write() [constructor]
//
// description:
//  construct an event_nasd_read object
//
// parameters:
//  copy_event - copy this object's event fields
//  start_time - start time for the event
//  write_param - parameters for the write request
//
event_nasd_write::event_nasd_write(
				   const event *copy_event,
				   simtime_t start_time,
				   param_write *write_param
				   ) : event_nasd(copy_event, start_time)
{
  parameter_ = write_param;
}

//
// event_nasd_write::~event_nasd_write() [destructor]
//
// description:
//  null destructor
//
// parameters:
//  none
//
event_nasd_write::~event_nasd_write()
{
}

//
// event_nasd_write::simulate()
//
// description:
//  simulate this event with the given parameters (from the comp parameter).
//  generates an event_nasd_write_complete class.
//
// parameters:
//  comp - a component object that contains all of the simulation parameters
//         for all components
//
// return values:
//  a pointer to a new event
//
#if defined(NASD_LOCAL_CACHE) && !defined(NASD_REMOTE_CACHING)
events_t *
event_nasd_write::simulate(
			   components &comp
			   )
{
  event *new_event;
  component_nasd *my_nasd;
  simtime_t completion_time;
  events_t *events = new events_t;
  block_list_t *block_list_local;
  size_t cache_size;

  my_nasd = comp.get_component_nasd(parameter_->get_nasd_id());

  completion_time = start_time_;
  // lookup the request in the local cache
  block_list_local =
    my_nasd->cache_lookup_local(parameter_->get_object(),
				parameter_->get_offset(),
				parameter_->get_length(),
				true, completion_time);

  if (block_list_local != NULL && block_list_local->size() != 0) {
    //
    // handle filling the cache
    //

    // winnow the block list down to the size of the cache if larger
    // than the cache.
    size_t block_size = my_nasd->get_block_size();
    size_t max_cache_size = my_nasd->get_max_cache_size();
    while ((block_list_local->size() * block_size) > max_cache_size) {
      block_list_local->pop_front(); // remove the oldest first
    }

    cache_size = my_nasd->get_cache_local_size();
    if (cache_size + block_list_local->size() * block_size > max_cache_size
	&& cache_size != 0) {
      block_list_t evicted_clean, evicted_dirty;

      // evict any entries based on the local disk reads
      my_nasd->cache_evict_local((cache_size + block_list_local->size() *
				  block_size - max_cache_size)/block_size,
				 parameter_->get_object(),
				 parameter_->get_offset(),
				 parameter_->get_length(),
				 evicted_clean,
				 evicted_dirty);

      // write out these entries
      if (!evicted_dirty.empty()) {
	completion_time += my_nasd->disk_io(&evicted_dirty, true,
					    completion_time);
      }
    }

    // add new entries to the cache
    completion_time += my_nasd->cache_add_local(block_list_local, true);
  }

  delete block_list_local;

  // update the perceived
  my_nasd->add_perceived_bw(completion_time - start_time_,
			    parameter_->get_length());
  // update request count
  my_nasd->inc_request_count();

  // update next request time
  my_nasd->set_next_request_time(completion_time);

  // generate response to request
  new_event = new event_WRITE_nasd_send_to_switch(dynamic_cast<event *>(this),
						  completion_time,
						  parameter_);
  events->push(new_event);

  return events;
}

#elif defined(NASD_LOCAL_CACHE) && defined(NASD_REMOTE_CACHING)

events_t *
event_nasd_write::simulate(
			   components &comp
			   )
{
  event *new_event;
  component_nasd *my_nasd;
  simtime_t completion_time;
  events_t *events = new events_t;
  block_list_t *block_list_local; // blocks not in the local cache
  nasd_global_map_t *global_map; // blocks that are in the global cache
  param_nasd_2_nasd *param;
  sim_bandwidth_t my_bandwidth;

  my_nasd = comp.get_component_nasd(parameter_->get_nasd_id());

  my_bandwidth = my_nasd->get_perceived_bw_cpy()[parameter_->get_nasd_id()].perceived_bw;

  size_t block_size = my_nasd->get_block_size();
  size_t max_cache_size = my_nasd->get_max_cache_size();

  // the simulator does not support requests that have offsets that are not on block aligned or
  // lengths that are block multiples
  ASSERT(parameter_->get_offset()%block_size == 0);
  ASSERT(parameter_->get_length()%block_size == 0);

  completion_time = start_time_;
  // lookup the request in the local cache
  block_list_local =
    my_nasd->cache_lookup_local(parameter_->get_object(),
				parameter_->get_offset(),
				parameter_->get_length(),
				true, completion_time);

  // if the local cache did not satisfy the request look it up in the
  // global cache
  if (block_list_local != NULL) {
    block_list_t tmp(*block_list_local);
    global_map =
      my_nasd->cache_lookup_remote(&tmp, true, completion_time);

    // generate the events for global reads
    if (global_map != NULL) {
      block_list_t tmp(*block_list_local);
      my_nasd->cache_remove_remote(tmp);

      for (nasd_global_map_iter_t iter = global_map->begin();
	   iter != global_map->end();
	   iter++) {
	param =
	  new param_parent_invalidate_request(parameter_->get_nasd_id(),
					      iter->first,
					      iter->second.size() *
					      sizeof(block_info_t),
					      new block_list_t(iter->second));
	new_event = new event_N2N_nasd_send_to_switch(completion_time,
						      injection_time_,
						      param);
	events->push(new_event);
      }

      delete global_map;
    }
  }

  // xxx make sure true's and false's are correct for write field of methods

  if (block_list_local != NULL && block_list_local->size() != 0) {

    //
    // handle filling the cache
    //
    if ((my_nasd->get_cache_size() + block_list_local->size() * block_size) <=
	max_cache_size) { // cache has enough room

      // read in the data that missed in the cache(s)
      completion_time += my_nasd->disk_io(block_list_local, false,
					  completion_time);

      // add new entries to the cache
      completion_time += my_nasd->cache_add_local(block_list_local, true);

    } else {
      // cache needs eviction

      block_list_t evicted_clean, evicted_dirty;
      int blocks_left = block_list_local->size();
      perceived_map_t perceived_bw_cpy = my_nasd->get_perceived_bw_cpy();
      vector<perceived_t> faster_nasds;
      sim_component_id_t local_eviction_nasd_dest = -1;
      
      // read in the data that missed in the cache(s)
      completion_time += my_nasd->disk_io(block_list_local, false,
					  completion_time);
      
      // make sure we don't have more data than the local and global cache
      // can possibly hold
      while ((block_list_local->size() * block_size) >
	     max_cache_size * perceived_bw_cpy.size()) {
#if defined(DEBUG)
	cout << "removing: block = " << block_list_local->front().block_id_;
#endif
	block_list_local->pop_front(); // remove the oldest first
      }

      // find the highest bandwidth NASD
      create_sorted_perceived(&perceived_bw_cpy, parameter_->get_nasd_id(),
			      faster_nasds);
ASSERT(faster_nasds.size() > 0);

      //
      // Evict from "local" cache all that needs to be evicted
      // this cache includes both entries from the private cache and the
      // global cache
      //
      do {
	if (faster_nasds.back().nasd_id == parameter_->get_nasd_id()) {
	  // handle eviction on local
	  // evict any entries based on the local disk reads
	  my_nasd->cache_evict_local(blocks_left,
				     parameter_->get_object(),
				     parameter_->get_offset(),
				     parameter_->get_length(),
				     evicted_clean,
				     evicted_dirty);
	  
	  blocks_left -= evicted_clean.size() + evicted_dirty.size();
	  
	  // write out these entries
	  if (!evicted_dirty.empty()) {
	    completion_time += my_nasd->disk_io(&evicted_dirty, true,
						completion_time);
	  }

	  // send these entries to a faster NASD
	  perceived_bw_cpy = my_nasd->get_perceived_bw_cpy();
	  vector<perceived_t> faster_nasds_tmp;
	  create_sorted_perceived(&perceived_bw_cpy, parameter_->get_nasd_id(),
				  faster_nasds_tmp);
	  evicted_clean.splice(evicted_clean.begin(), evicted_dirty);
	  if (faster_nasds_tmp.back().nasd_id != parameter_->get_nasd_id() &&
	      fabs(faster_nasds_tmp.back().perceived_bw - my_bandwidth) > 5 && // changed this to fix oscillation
	      !evicted_clean.empty()) {
	    
	    local_eviction_nasd_dest = faster_nasds_tmp.back().nasd_id;

	    // add these blocks to remote table for future lookup
	    my_nasd->cache_add_remote(&evicted_clean,
				      faster_nasds_tmp.back().nasd_id);

	    param =
	      new param_parent_eviction_request(parameter_->get_nasd_id(),
						faster_nasds_tmp.back().nasd_id,
						evicted_clean.size() *
						block_size,
						new block_list_t(evicted_clean));
	    
	    // send these blocks to the global cache if a faster NASD is found
	    new_event = new event_N2N_nasd_send_to_switch(completion_time,
							  injection_time_,
							  param);
	    events->push(new_event);
	  }
	} else {
	  blocks_left =
	    my_nasd->cache_evict_global(blocks_left,
					faster_nasds.back().nasd_id);
	}

	faster_nasds.pop_back();
      } while (!faster_nasds.empty() && blocks_left > 0);
      
      //
      // find homes in the global cache for any of the blocks that do not
      // fit in the local cache
      //

      // ensure we don't overfill the cache
      while (block_list_local->size() * block_size + my_nasd->get_cache_size()
             > my_nasd->get_max_cache_size()) {
        block_list_local->pop_front();
      } 
      
      // fill the local cache
      if (block_list_local->size() < (max_cache_size/block_size)) {
	completion_time += my_nasd->cache_add_local(block_list_local, true);
	blocks_left = 0;
      } else {
	size_t copied_num;
	block_list_t *tmp = new block_list_t;
	
	for (copied_num = 0;
	     copied_num < (max_cache_size/block_size); copied_num++) {
	  tmp->push_back(block_list_local->front());
	  block_list_local->pop_front();
	}
	
	completion_time += my_nasd->cache_add_local(tmp, false);
	
	blocks_left -= max_cache_size/block_size;
      }

      //
      // find homes for these evicted blocks
      //
      // find the highest bandwidth NASD that has cache entries
      perceived_bw_cpy = my_nasd->get_perceived_bw_cpy();
      create_sorted_perceived(&perceived_bw_cpy, parameter_->get_nasd_id(),
			      faster_nasds);
      while (blocks_left > 0 && !faster_nasds.empty()) {
	if (faster_nasds.back().nasd_id != parameter_->get_nasd_id() &&
	    fabs(faster_nasds.back().perceived_bw - my_bandwidth) > 5 && // changed this to fix oscillation
	    faster_nasds.back().nasd_id != local_eviction_nasd_dest) {
	  block_list_t *remote_blocks = new block_list_t;
	  
	  ASSERT(block_list_local->size() == blocks_left);
	  
	  // add the blocks to remote_blocks
	  while (remote_blocks->size() < (max_cache_size / block_size) &&
		 blocks_left > 0) {
	    remote_blocks->push_back(block_list_local->front());
	    block_list_local->pop_front();
	    blocks_left--;
	  }
	  
	  // add these blocks to remote table for future lookup
	  my_nasd->cache_add_remote(remote_blocks,
				    faster_nasds.back().nasd_id);

	  param = 
	    new param_parent_eviction_request(parameter_->get_nasd_id(),
					      faster_nasds.back().nasd_id,
					      remote_blocks->size() *
					      block_size,
					      remote_blocks);
	  
	  // send these blocks to the global cache if a faster NASD is found
	  new_event = new event_N2N_nasd_send_to_switch(completion_time,
							injection_time_,
							param);
	  events->push(new_event);
	}

	faster_nasds.pop_back();
      }
    }

  }

  delete block_list_local;

  // generate response to request
  new_event =
    new event_WRITE_nasd_send_to_switch(dynamic_cast<event *>(this),
					completion_time, parameter_);
  events->push(new_event);

  // update the perceived
  my_nasd->add_perceived_bw(completion_time - start_time_,
			    parameter_->get_length());
  // update request count
  my_nasd->inc_request_count();

  // update next request time
  my_nasd->set_next_request_time(completion_time);

  return events;
}
#endif

#if defined(DEBUG)
//
// event_nasd_write::print()
//
// description:
//  When DEBUG is enabled, this method display information about the event.
//
// parameters:
//  output - the ostream for display
//
// return values:
//  none
// 
void
event_nasd_write::print(
			ostream &output
			) const
{
  output << "nasd_write: { start = " << start_time_ << ", ";
  parameter_->print(output);
  output << " }";
}
#endif // defined(DEBUG)

#if defined(NASD_REMOTE_CACHING)
//
// event_nasd_N2N
//
event_nasd_N2N::event_nasd_N2N(
			       const event *copy_event,
			       simtime_t start_time,
			       param_nasd_2_nasd *parameter
			       )
  : event_nasd(copy_event, start_time)
{
  parameter_ = parameter;
}

event_nasd_N2N::~event_nasd_N2N()
{
}

events_t *
event_nasd_N2N::simulate(
			 components &comp
			 )
{ // xxx fix this, figure out where to delete block_list from parameter_
  events_t *events;
  
  if (dynamic_cast<param_read_forward_request *>(parameter_) != NULL) {
    // request is for a forwarded read
    events = global_cache_read(comp);    
  } else if (dynamic_cast<param_parent_eviction_request *>(parameter_)
	     != NULL) {
    // request is for this nasd to store blocks in the global cache
    global_cache_add(comp);
    events = new events_t;
  } else if (dynamic_cast<param_parent_invalidate_request *>(parameter_)
	     != NULL) {
    // request is for this nasd to invalidate blocks in its global cache
    global_cache_invalidate(comp);
    events = new events_t;
  } else if (dynamic_cast<param_returned_read_request *>(parameter_)
	     != NULL) {
    // This is pretty much a hack to get returned reads to work. Maybe
    // there is a more elegant want to do this, but it isn't too bad.
    // generate_local_reads() creates a set of event_nasd_reads
    // with the parameter's partial set correctly. event_nasd_read::simulate()
    // understands these and updates its information correctly to satisfy
    // the partial request(s) locally, evicting cache entires if needed.
    events = generate_local_reads(comp);
  }

  return events;
}

void
event_nasd_N2N::global_cache_add(
				 components& comp
				 )
{
  sim_component_id_t my_nasd_id = parameter_->get_destination_nasd();
  component_nasd *my_nasd =
    comp.get_component_nasd(my_nasd_id);
  block_list_t *block_list =
    dynamic_cast<param_parent_eviction_request *>(parameter_)
    ->get_block_list();
  perceived_map_t perceived_bw_cpy;
  perceived_map_t::iterator iter, highest;
  int needed_blocks = block_list->size();
  block_list_t local_dirty, local_clean;
  simtime_t completion_time = my_nasd->get_next_request_time();

  ASSERT(needed_blocks * my_nasd->get_block_size() <=
	 my_nasd->get_max_cache_size());

  if (completion_time < start_time_) {
    completion_time = start_time_;
  }

  // evict if needed
  if (my_nasd->get_cache_size() +
      needed_blocks * my_nasd->get_block_size() >
      my_nasd->get_max_cache_size()) {
    // start with the highest b/w NASD and evict from there

    // set up the perceived
    perceived_bw_cpy = my_nasd->get_perceived_bw_cpy();

    while (needed_blocks > 0) {
      for (iter = perceived_bw_cpy.begin(), highest = iter;
	   iter != perceived_bw_cpy.end();
	   iter++) {
	if (iter->second.perceived_bw > highest->second.perceived_bw) {
	  highest = iter;
	}
      }

      if (highest->first == my_nasd_id) {
	// evict
	my_nasd->cache_evict_local(needed_blocks,
				   0/* don't care */,
				   0/* don't care */, 0 /* don't care */,
				   local_clean, local_dirty);
	needed_blocks -= local_clean.size() + local_dirty.size();

	// write out dirty evicted cache entries
	if (!local_dirty.empty()) {
	  completion_time += my_nasd->disk_io(&local_dirty, true,
					      completion_time);

	  my_nasd->set_next_request_time(completion_time);
	}
      } else {
	needed_blocks = my_nasd->cache_evict_global(needed_blocks,
						    highest->first);
      }

      perceived_bw_cpy.erase(highest);
    }
  }

  // add
  my_nasd->cache_add_global(block_list, parameter_->get_source_nasd());

  delete block_list;
  delete parameter_;
}

events_t *
event_nasd_N2N::global_cache_read(
				  components& comp
				  )
{
  component_nasd *my_nasd =
    comp.get_component_nasd(parameter_->get_destination_nasd());
  block_list_t *sat_block_list, *unsat_block_list;
  param_read_forward_request *param =
    dynamic_cast<param_read_forward_request *>(parameter_);
  simtime_t completion_time = my_nasd->get_next_request_time();
  event *new_event;
  events_t *events;

  events = new events_t;

  if (completion_time < start_time_) {
    completion_time = start_time_;
  }
  unsat_block_list = param->get_block_list();
  sat_block_list =
    my_nasd->cache_lookup_global(unsat_block_list,
				 param->get_source_nasd(),
				 completion_time);

  // generate response to client for satisfied blocks
  if (!sat_block_list->empty()) {
    param->get_param_read()->set_length(sat_block_list->size() *
					my_nasd->get_block_size());
    // only set these if we are fragmenting the request
    if (!unsat_block_list->empty()) {
      param->get_param_read()->set_partial(true);
    }

    new_event = new event_READ_nasd_send_to_switch(dynamic_cast<event *>(this),
						   completion_time,
						   param->get_param_read());
    events->push(new_event);
  }

  // generate new read for unsatisfied blocks. this will go to the owner
  // NASD
  if (!unsat_block_list->empty()) {
    param_returned_read_request *param_new = new
      param_returned_read_request(param->get_destination_nasd(), // invert these since the request is being bounced back
				  param->get_source_nasd(),
				  unsat_block_list->size() *
				  sizeof(block_info_t),
				  unsat_block_list,
				  param->get_param_read());
    new_event = new event_N2N_nasd_send_to_switch(completion_time,
						  injection_time_,
						  param_new);

    events->push(new_event);
#if defined(DEBUG)
    cout << "event_nasd_N2N::global_cache_read(): returning request" << endl;
#endif
  } else {
    delete unsat_block_list;
  }

  delete sat_block_list;
  delete param; // should not destroy the parameter contained within

  return events;
}

void
event_nasd_N2N::global_cache_invalidate(
					components& comp
					)
{
  component_nasd *my_nasd =
    comp.get_component_nasd(parameter_->get_destination_nasd());
  param_parent_invalidate_request *param =
    dynamic_cast<param_parent_invalidate_request *>(parameter_);

  my_nasd->cache_invalidate_global(param->get_block_list(),
				   param->get_source_nasd());

  delete param->get_block_list();
  delete param;
}

events_t *
event_nasd_N2N::generate_local_reads(
				     components &comp
				     )
{
  block_list_t *block_list;
  events_t *events;
  event *event;
  sim_offset_t prev_offset, start, first;
  param_read *my_param_read;
  param_returned_read_request *param =
    dynamic_cast<param_returned_read_request *>(parameter_);
  size_t block_size =
    comp.get_component_nasd(param->get_source_nasd())->get_block_size();

  block_list = param->get_block_list();

  events = new events_t;

  ASSERT(!block_list->empty());

  for (prev_offset = block_list->front().block_id_, start = prev_offset,
	 first = start, block_list->pop_front();
       !block_list->empty();
       prev_offset = block_list->front().block_id_, block_list->pop_front()) {
    if (prev_offset + block_size != block_list->front().block_id_) {
      my_param_read = new param_read(*param->get_param_read());
      my_param_read->set_partial(true);
      my_param_read->set_returned(true);
      my_param_read->set_length(prev_offset + block_size - start);

      event = new event_nasd_read(this, start_time_, my_param_read);
      events->push(event);

      start = prev_offset;
    }
  }

  // generate the last request
  my_param_read = new param_read(*param->get_param_read());
  if (first != start) { // this request was fragmented
    my_param_read->set_partial(true);
  }
  my_param_read->set_returned(true);
  my_param_read->set_length(prev_offset + block_size - start);

  event = new event_nasd_read(this, start_time_, my_param_read);
  events->push(event);

  delete block_list;
  delete param;

  return events;
}

#if defined(DEBUG)
void
event_nasd_N2N::print(
		      ostream &output
		      ) const
{
  output << "nasd_N2N: { start = " << start_time_ << ", ";
  if (dynamic_cast<param_read_forward_request *>(parameter_) != NULL) {
    dynamic_cast<param_read_forward_request *>(parameter_)->print(output);
  } else if (dynamic_cast<param_parent_eviction_request *>(parameter_)
	     != NULL) {
    dynamic_cast<param_parent_eviction_request *>(parameter_)->print(output);
  } else if (dynamic_cast<param_parent_invalidate_request *>(parameter_)
	     != NULL) {
    dynamic_cast<param_parent_invalidate_request *>(parameter_)->print(output);
  }
  output << " }";
}
#endif // defined(DEBUG)

#endif // defined(NASD_REMOTE_CACHING)
