// $Id: tpcb.C,v 1.1 2000/05/09 13:47:30 ozaki Exp $

// This is the generator that tries to model a TPC-B benchmark. However, we have
// made a number of assumptions. Under the time constraints this is probably the most
// we can do.
//
// Check http://www.tpc.org/bspec.html for the real thing!

#include <fstream>
#include <iomanip>
#include <stdlib.h>
#include <time.h>
#include <limits.h>

#define READ     0
#define WRITE    1

#define BRANCHES 0
#define TELLERS  1
#define ACCOUNTS 2
#define HISTORY  3

struct relInfo
{
  unsigned long numRec;
  unsigned int  recLen;
  unsigned long fileSize;
};

relInfo relations [4];

// prototypes
void dump_read (ofstream & fout, float currTime, int objId, int offset, int size, 
		int nasdId, long streamId);
void dump_write (ofstream & fout, float currTime, int objId, int offset, int size, 
		 int nasdId, long streamId);
void doIo (ofstream & fout, int val, int recLen, int recNum, 
	   float & currTime, int type);
void doIoFile (ofstream & fout, int val, int recLen, int numRec, 
	       float & currTime, int type);
void createDatabase (ofstream &fout, float &currTime);
void DoOne (ofstream &fout, float & currTime, long Bid, long Tid, long  Aid);

unsigned int numNasds    = 5;
unsigned int stripeSize  = 10240;
float        reqGap      = 0.001;
long         stream_id   = 0;

long tps                 =      50;  /* the tps scaling factor: here it is 1 */
long nbranches           =       1;  /* number of branches in 1 tps db       */
long ntellers            =      10;  /* number of tellers in  1 tps db       */
long naccounts           =  100000;  /* number of accounts in 1 tps db       */
long nhistory            =  864000;  /* number of history recs in 1 tps db   */
int histIndex            =       0;
unsigned long long uniqueBytes = 0;
 
int 
main(int argc, char **argv)
{
  if (argc != 2) {
    cerr << "Usage: tpcb <trace file name>" << endl;
    exit (-1);
  }
  ofstream fout (argv [1]);
  if (!fout) { exit (-1); }
    
  float startTime       = 0;
  float currTime        = startTime;
  
  createDatabase (fout, currTime);
  
  // random number generator
  srandom (time (0));
  
  // Do transactions
  for (int i = 0; i < tps; i++) {
    
    // The methodology we are using now to get the Bid, Tid and Aid is not how 
    // it is supposed to be done in the real TPCB. In TPCB even though the choice
    // is random but, it is based on certain well defined distributions. If we have
    // time we will do it that way.
    
    long Bid = random () % nbranches;
    long Tid = random () % ntellers;
    long Aid = random () % naccounts;
    
    DoOne (fout, currTime, Bid, Tid, Aid);
  }
  
  cout << "WORKING SET SIZE = " << uniqueBytes << endl;
}

/*
 *  DoOne - Executes a single TPC BM B transaction.
 *  Basically do a read / modify / write
 */
void
DoOne (ofstream &fout, float & currTime, long Bid, long Tid, long  Aid)
{
        
  //exec sql BEGIN WORK;      
  
  //exec sql UPDATE accounts  
  //    SET             Abalance = Abalance + :delta
  //    WHERE           Aid = :Aid;
  doIo (fout, 
	ACCOUNTS, 
	relations [ACCOUNTS].recLen, 
	Aid,
	currTime,
	READ);
  doIo (fout, 
	ACCOUNTS, 
	relations [ACCOUNTS].recLen, 
	Aid,
	currTime,
	WRITE);
  
  //exec sql SELECT Abalance INTO :Abalance
  //    FROM            accounts
  //    WHERE           Aid = :Aid;
  doIo (fout, 
	ACCOUNTS, 
	relations [ACCOUNTS].recLen, 
	Aid,
	currTime,
	READ);
  
  // exec sql UPDATE tellers  
  //    SET             Tbalance = Tbalance + :delta
  //    WHERE           Tid = :Tid;
  doIo (fout, 
	TELLERS, 
	relations [TELLERS].recLen, 
	Tid,
	currTime,
	READ);
  doIo (fout, 
	TELLERS, 
	relations [TELLERS].recLen, 
	Tid,
	currTime,
	WRITE);
  
  //exec sql UPDATE branches  
  //    SET             Bbalance = Bbalance + :delta
  //    WHERE           Bid = :Bid;
  doIo (fout, 
	BRANCHES,
	relations [BRANCHES].recLen, 
	Bid,
	currTime,
	READ);
  doIo (fout, 
	BRANCHES, 
	relations [BRANCHES].recLen, 
	Bid,
	currTime,
	WRITE);
  
  //exec sql INSERT INTO history(Tid, Bid, Aid, delta, time) 
  //    VALUES (:Tid, :Bid, :Aid, :delta, CURRENT);
  
  if (histIndex >= nhistory) {
    cerr << "ASSERTION failed: (histIndex >= nhistory) " << endl;
    exit (-1);
  }

  doIo (fout, 
	HISTORY, 
	relations [HISTORY].recLen, 
	histIndex,
	currTime,
	WRITE);
  histIndex++;

  uniqueBytes += relations [HISTORY].recLen; // Add to working set
  
  //exec sql COMMIT WORK;
  
}                       /* end of DoOne                 */

void createDatabase (ofstream &fout, float &currTime) 
{
  relations [BRANCHES].numRec   = nbranches;
  relations [BRANCHES].recLen   = 100;
  relations [BRANCHES].fileSize = nbranches * 100;
  
  doIoFile (fout, BRANCHES, 100, nbranches, currTime, WRITE);

  relations [TELLERS].numRec    = ntellers;
  relations [TELLERS].recLen    = 100;
  relations [TELLERS].fileSize  = ntellers * 100;
  
  doIoFile (fout, TELLERS, 100, ntellers, currTime, WRITE);

  relations [ACCOUNTS].numRec   = naccounts;
  relations [ACCOUNTS].recLen   = 100;
  relations [ACCOUNTS].fileSize = naccounts * 100;
  
  doIoFile (fout, ACCOUNTS, 100, naccounts, currTime, WRITE);  
  
  relations [HISTORY].numRec    = nhistory;
  relations [HISTORY].recLen    = 100;
  relations [HISTORY].fileSize  = nhistory * 100;
  
  uniqueBytes = 
    relations [BRANCHES].fileSize +
    relations [TELLERS].fileSize +
    relations [ACCOUNTS].fileSize +
    relations [HISTORY].fileSize;
  
  // I am not taking the history file into account because it is just written
}

void
doIoFile (
	  ofstream & fout,
	  int val,
	  int recLen,
	  int numRec,
	  float &currTime,
	  int type
	  )
{
  long size = recLen * numRec;
  
  int stOid          = val * 2000;
  int endOid         = ((val + 1) * 2000) - 1;
  int currOid        = stOid;
  
  bool start         = true;
  int  i;
  
  while (size > 0) 
    {
      if (start) {
	start = false; i = val % numNasds;
      }
      else i = 0;
      
      for (; (i < numNasds) && (size > 0); i++) 
	{
	  if (currOid > endOid) {
	    cerr << "ASSERTION failed: (currOid < endOid + 1)" << endl;
	    exit (-1);
	  }
	  
	  int s = (size > stripeSize)? stripeSize : size;
	  
	  if (type == READ)
	    dump_read (fout, currTime, currOid, 0, s, i, stream_id); 
	  else
	    dump_write (fout, currTime, currOid, 0, s, i, stream_id); 

	  fout << endl;
	  currTime += reqGap;
	  
	  size = size - stripeSize;
	}
      
      currOid++;
    }
    
    stream_id++; 
    
    // wrap around
    if (stream_id == LONG_MAX) stream_id = 0;
}

// read or write a record
void
doIo (
      ofstream & fout, 
      int val,                  // fileId
      int recLen,               // length of each record
      int recNum,               // record number 
      float & currTime,         // current time
      int type
      ) 
{
  long logicalOffset = recLen * recNum;
  
  // chunk # with respect to 1st chunk
  int n              = logicalOffset / stripeSize;              // integer division
  int nasdId         = ((val % numNasds) + n) % numNasds;
  int stripeLevel    = logicalOffset / (stripeSize * numNasds); // integer division
  int stOid          = val * 2000 + stripeLevel;
  int endOid         = ((val + 1) * 2000) - 1;
  int currOid        = stOid;
  int startOffset    = logicalOffset - n * stripeSize;
  int size           = recLen;
  int startSize      = stripeSize - startOffset;
  
  if (startSize > size) startSize = size;
  
  bool start         = true;
  int  i;
    
  while (size > 0) 
    {
      if (start) {
	i = nasdId;
      }
      else i = 0;
      
      for (; (i < numNasds) && (size > 0); i++) 
	{
	  if (currOid > endOid) {
	    cerr << "ASSERTION failed: (currOid < endOid + 1)" << endl;
	    exit (-1);
	  }
	  
	  int s, o;
	  
	  if (!start) {
	    s = (size > stripeSize)? stripeSize : size;
	    o = 0;
	  }
	  else {
	    s = startSize;
	    o = startOffset;
	  }
	  
	  if (type == READ)
	    dump_read (fout, currTime, currOid, o, s, i, stream_id); 
	  else
	    dump_write (fout, currTime, currOid, o, s, i, stream_id); 
	  
	  fout << endl;
	  currTime += reqGap;
	  
	  size = size - s;
	  start = false;
	}
      
      currOid++;
    }
  
  stream_id++; 
    
  // wrap around
  if (stream_id == LONG_MAX) stream_id = 0;
}  

void 
dump_read (ofstream & fout, float currTime, int objId, int offset, int size, 
	   int nasdId, long streamId)
{
  fout << currTime << " read " << objId << " " << offset << " " << size << " "
       << nasdId << " " << streamId;
}

void 
dump_write (ofstream & fout, float currTime, int objId, int offset, int size, 
	    int nasdId, long streamId)
{
  fout << currTime << " write " << objId << " " << offset << " " << size << " "
       << nasdId << " " << streamId;
}


