// $Id: traceWeb.C,v 1.6 2000/05/10 21:52:35 ozaki Exp $

#include <fstream>
#include <iomanip>
#include <stdlib.h>
#include <time.h>
#include <assert.h>
#include <limits.h>
#include <math.h>

// Creates synthetic data set to model reads for a web server

// prototypes
void readFile (ofstream & fout, const int fileId, const unsigned fileSize, 
	       const int numNasds, float & currTime, const int stripeSize,
	       const long streamId);
void dump_read (ofstream & fout, float currTime, int objId, int offset, int size, 
		int nasdId, long streamId);

float reqGap;

int 
main(int argc, char **argv)
{
  if (argc != 2) {
    cerr << "Usage: genWeb <trace file name>" << endl;
    exit (-1);
  }
  
  ofstream fout (argv [1]);
  if (!fout) { exit (-1); }
  
  //
  // Initialize synthetic generator's paramters
  //
  unsigned int numNasds    = 5;
  unsigned int stripeSize  = 10240;
  unsigned int numFileReq  = 1000;
  unsigned int numFiles    = 250;
  float        startTime   = 0;
  reqGap                   = 0.001;
  
  unsigned int htmlSize    = 12288;
  unsigned int imageSize   = 10240;
#if 0
  unsigned int soundSize   = 199680;
  unsigned int videoSize   = 593920;
#else
  unsigned int soundSize   = 12288;
  unsigned int videoSize   = 10240;
#endif
  float        htmlpc      = 49;
  float        imagepc     = 48;
  float        soundpc     = 2;
  float        videopc     = 1;
  
  // each request for a file will be an independent request. However, the
  // sub-requests for the objects of the file (since the file is striped) will be
  // dependent on each other. Each request for a file will be allocated a
  // monotonically increasing streamId that all its sub-requests will use
  long         streamId    = 0;
  
  // End of parameters for the generator ------------------------------------------
  
  // Partition [0 .. numFiles - 1] into html, image, sound & video categories
  unsigned int numHtml     = (unsigned int) floor ((numFiles * htmlpc) / 100.0);
  unsigned int numImage    = (unsigned int) floor ((numFiles * imagepc) / 100.0);
  unsigned int numSound    = (unsigned int) floor ((numFiles * soundpc) / 100.0);
  unsigned int numVideo    = (unsigned int) floor ((numFiles * videopc) / 100.0);
  
  assert ((numHtml + numImage + numSound + numVideo) <= numFiles);
  
  unsigned int htmlStNum   = 0;
  unsigned int imageStNum  = numHtml;
  unsigned int soundStNum  = imageStNum + numImage;
  unsigned int videoStNum  = soundStNum + numSound;
  
  // Set current time
  float currTime           = startTime;
  
  // stats
  unsigned int imageCnt    = 0;
  unsigned int soundCnt    = 0;
  unsigned int htmlCnt     = 0;
  unsigned int videoCnt    = 0;
  unsigned long long uniqueBytes = 0;
  
  // number of files is numFiles. So make an array of boolean that will keep
  // track whether this file was referenced before or not.
  bool * fileRef           =  new bool [numFiles];
  for (int i = 0; i < numFiles; i++) fileRef [i] = false;
  
  // random number generator
  srandom (time (0));
  
  // Main request generation loop
  for (int i = 0; i < numFileReq; i++) 
    {
      int  fileSize;
      long val = random ();
      val %= numFiles;
      
      if (val < imageStNum) { 
	htmlCnt ++; 
	readFile (fout, val, htmlSize, numNasds, currTime, stripeSize, streamId); 
	fileSize = htmlSize;
      }
      else if (val < soundStNum) {
	imageCnt ++; 
	readFile (fout, val, imageSize, numNasds, currTime, stripeSize, streamId);
	fileSize = imageSize;
      }
      else if (val < videoStNum) {
	soundCnt ++; 
	readFile (fout, val, soundSize, numNasds, currTime, stripeSize, streamId);
	fileSize = soundSize;
      }
      else {
	videoCnt ++; 
	readFile (fout, val, videoSize, numNasds, currTime, stripeSize, streamId);
	fileSize = videoSize;
      }    
      
      // Update uniquebytes info
      if (!fileRef [val]) { 
	fileRef [val] = true; uniqueBytes += fileSize;
      }
      
      // Update streamId. Wrap around if needed
      streamId++; if (streamId == LONG_MAX) streamId = 0;
    }
  
  // Dump out stats
  cout << "HTML:  NumFiles = " << numHtml << ", NumRequests = " << htmlCnt 
       << ", % of total = " << (htmlCnt * 100) / (float)numFileReq
       << endl
       << "IMAGE: NumFiles = " << numImage << ", NumRequests = " << imageCnt 
       << ", % of total = " << (imageCnt * 100) / (float)numFileReq
       << endl
       << "SOUND: NumFiles = " << numSound << ", NumRequests = " << soundCnt 
       << ", % of total = " << (soundCnt * 100) / (float)numFileReq
       << endl
       << "VIDEO: NumFiles = " << numVideo << ", NumRequests = " << videoCnt 
       << ", % of total = " << (videoCnt * 100) / (float)numFileReq
       << endl << endl
       << "Unique Bytes touched = " << uniqueBytes << endl;
  
  delete fileRef;
  
  return 0;
}    

void 
readFile (ofstream & fout, const int fileId, const unsigned fileSize, 
	  const int numNasds, float & currTime, const int stripeSize,
	  const long streamId) 
{
  int size                 = fileSize;
  
  unsigned int stOid       = fileId * 100;
  unsigned int endOid      = ((fileId + 1) * 100) - 1;
  unsigned int currOid     = stOid;
  
  // A simple hack to make striping occur starting at different nasds for files with
  // different ids. The idea is that the workload is better spaced out over the
  // NASDs
  
  bool start = true;
  int  i;
  
  while (size > 0) 
    {
      if (start) {
	start = false; i = fileId % numNasds;
      }
      else i = 0;
	
      for (; (i < numNasds) && (size > 0); i++) 
	{
	  assert (currOid < endOid + 1);
		    
	  int s = (size > stripeSize)? stripeSize : size;
	    
	  dump_read (fout, currTime, currOid, 0, s, i, streamId); fout << endl;
	  currTime += reqGap;
	  
	  size = size - stripeSize;
	}
      
      currOid++;
    }
}

void 
dump_read (ofstream & fout, float currTime, int objId, int offset, int size, 
	   int nasdId, long streamId)
{
  fout << currTime << " read " << objId << " " << offset << " " << size << " "
       << nasdId << " " << streamId;
}
